// Copyright(c) 2018, Intel Corporation
//
// Redistribution  and  use  in source  and  binary  forms,  with  or  without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of  source code  must retain the  above copyright notice,
//   this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// * Neither the name  of Intel Corporation  nor the names of its contributors
//   may be used to  endorse or promote  products derived  from this  software
//   without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,  BUT NOT LIMITED TO,  THE
// IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT  SHALL THE COPYRIGHT OWNER  OR CONTRIBUTORS BE
// LIABLE  FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR
// CONSEQUENTIAL  DAMAGES  (INCLUDING,  BUT  NOT LIMITED  TO,  PROCUREMENT  OF
// SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE,  DATA, OR PROFITS;  OR BUSINESS
// INTERRUPTION)  HOWEVER CAUSED  AND ON ANY THEORY  OF LIABILITY,  WHETHER IN
// CONTRACT,  STRICT LIABILITY,  OR TORT  (INCLUDING NEGLIGENCE  OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

#include <getopt.h>
#include "fpgainfo.h"
#include "sysinfo.h"
#include "phyinfo.h"
#include "bmcinfo.h"
#include <opae/fpga.h>
#include <wchar.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include "safe_string/safe_string.h"

static int num_ports = 8;
static int speed = 10;

/*
 * phy command configuration, set during parse_phy_args()
 */
static struct _phy_config {
	int group_num;
	int show_pkvl;
} phy_config = {.group_num = -1,
				.show_pkvl = 1};

/*
 * Print help
 */
void phy_help(void)
{
	printf("\nPrint PHY information\n"
		   "        fpgainfo phy [-h] [-G <group-number>]\n"
		   "                -h,--help           Print this help\n"
		   "                -G,--group          Select PHY group {0,1,all}\n"
		   "\n");
}

#define PATH1_KEYWORD   "pac"
#define PATH2_KEYWORD   "misc"
#define PHYGRP_DEVNAME  "eth_group"
#define CDEV_ID_SIZE    8
#define FPGA_PHY_GROUP_GET_INFO 0xB702
static bool get_eth_group_path(const char *in_path, int pac_net_index,
							   int group_num, char *out_path, int size)
{
	DIR *dir = NULL;
	struct dirent *dirent = NULL;
	char path[SYSFS_PATH_MAX] = {0};
	char *p;
	char *substr;
	int i = 0;
	int pac_net_found_count = 0;
	int result;
	bool found = false;

	if (in_path == NULL || out_path == NULL)
		return found;
	strncpy_s(path, sizeof(path), in_path, strlen(in_path));

	while (1) {
		dir = opendir(path);
		if (NULL == dir) {
			break;
		}
		while (NULL != (dirent = readdir(dir))) {
			if (EOK == strcmp_s(dirent->d_name, strlen(dirent->d_name),
								".", &result)) {
				if (result == 0)
					continue;
			}
			if (EOK == strcmp_s(dirent->d_name, strlen(dirent->d_name),
								"..", &result)) {
				if (result == 0)
					continue;
			}
			if (EOK == strstr_s(dirent->d_name, strlen(dirent->d_name),
								PATH2_KEYWORD, strlen(PATH2_KEYWORD),
								&substr)) {
				snprintf_s_s(path+strlen(path), sizeof(path)-strlen(path),
							 			 "/%s", dirent->d_name);
				break;
			}
			if (EOK == strstr_s(dirent->d_name, strlen(dirent->d_name),
								PATH1_KEYWORD, strlen(PATH1_KEYWORD),
								&substr)) {
				if (pac_net_found_count == pac_net_index) {
					snprintf_s_s(path+strlen(path), sizeof(path)-strlen(path),
							 	 "/%s", dirent->d_name);
					break;
				} else {
					pac_net_found_count += 1;
					continue;
				}
			}

			if (!strncmp(dirent->d_name, PHYGRP_DEVNAME,
				 strlen(PHYGRP_DEVNAME))) {
				p = strrchr(dirent->d_name, '.');
				if (NULL == p) {
					fprintf(stderr, "Invalid eth_group name %s\n",
							dirent->d_name);
					continue;
				}
				++p;
				sscanf_s_i(p, "%d", &i);
				if (group_num == i) {
					snprintf_s_ss(out_path, size, "%s/%s/dev", path,
								  dirent->d_name);
					found = true;
					break;
				}
			}
		}
		closedir(dir);
		if (dirent == NULL || found)
			break;
	}
	return found;
}

static void print_phy_group_info(fpga_properties props, int group_num)
{
	char path[SYSFS_PATH_MAX];
	const char *sysfspath = get_sysfs_path(props, FPGA_DEVICE, NULL);
	bool found = false;
	int fd;
	ssize_t sz;
	char cdevid[CDEV_ID_SIZE];

	printf("//****** PHY GROUP %d ******//\n", group_num);

	// Search specified PHY group
	found = get_eth_group_path(sysfspath, 0, group_num, path, SYSFS_PATH_MAX);
	if (!found) {
		found = get_eth_group_path(sysfspath, 1, group_num, path,
								   SYSFS_PATH_MAX);
	}

	if (found) {
		// Get char device ID from PHY group sysfs device
		fd = open(path, O_RDONLY);
		if (fd < 0) {
			fprintf(stderr, "Open %s failed\n", path);
			return;
		}
		sz = read(fd, cdevid, CDEV_ID_SIZE);
		if (sz > 0)
			cdevid[sz-1] = '\0';
		close(fd);

		// Get information from PHY group char device
		snprintf_s_s(path, sizeof(path), "/dev/char/%s", cdevid);
		fd = open(path, O_RDWR);
		if (fd < 0) {
			fprintf(stderr, "Open %s failed\n", path);
			return;
		}
		struct fpga_phy_group_info {
			unsigned int    argsz;
			unsigned int    flags;
			unsigned char  speed;
			unsigned char  phy_num;
			unsigned char  mac_num;
			unsigned char  group_id;
		} info = {
			.argsz = 14,
			.flags = 0,
			.speed = 0,
			.phy_num = 0,
			.mac_num = 0,
			.group_id = 0
		};
		if (0 == ioctl(fd, FPGA_PHY_GROUP_GET_INFO, &info)) {
			printf("%-29s : %s\n", "Direction",
				   info.group_id == 0 ? "Line side" : "Host side");
			printf("%-29s : %d Gbps\n", "Speed", info.speed);
			printf("%-29s : %d\n", "Number of PHYs", info.phy_num);
		} else {
			printf("ioctl error\n");
		}
		if (info.group_id == 0) {
			num_ports = info.phy_num;
			speed = info.speed;
		}
		close(fd);
	} else {
		fprintf(stderr, "WARNING: eth_group %d not found\n", group_num);
	}
}

#define PKVL_NAME	"pkvl"
#define SPI_NAME	"spi"
#define MAX_PORTS	8

static void print_pkvl_info(fpga_properties props)
{
	char path[SYSFS_PATH_MAX] = {0};
	struct dev_list *lptr = NULL;
	const char *sysfspath = get_sysfs_path(props, FPGA_DEVICE, &lptr);
	int offset;
	int mask = 0;
	int fpga_mode = 0 ;
	int result;
	int i,j;
	char mode[16] = {0};
	char status[16] = {0};
	char version[16] = {0};

	printf("//****** Intel C827 Retimer ******//\n");

	fpga_mode = (lptr->fme->fpga_bitstream_id >> 32)&0xf;

	if (sysfspath == NULL)
		return;
	snprintf_s_s(path, sizeof(path),
				 "%s/spi-altera.*.auto/spi_master/spi*/spi*.*/pkvl",
				 sysfspath);
	if (glob_sysfs_path(path) != FPGA_OK) {
		fprintf(stderr, "WARNING: pkvl not found\n");
		return;
	}

	offset = strlen(path);
	snprintf_s_s(path+offset, sizeof(path)-offset, "/%s", "status");
	get_sysfs_attr(path, status, sizeof(status));
	result = strtol(status, NULL, 16);
	if (speed==10) {
		/* 8x10g */
		mask = 0xff;
	} else if (speed == 25) {
		switch(fpga_mode) {
		case 1: /* 4x25g */
		case 3: /* 6x25g */
			mask = 0xf;
			break;
		case 2: /* 2x1x25g */
		case 5: /* 2x1x25gx2FVL */
			mask = 0x11;
			break;
		case 4: /* 2x2x25g */
			mask = 0x33;
			break;
		case 6: /* 1x2x25g */
			mask = 0x03;
			break;
		default:
			mask = 0x0;
			break;
		}
	}
	strncpy_s(mode, sizeof(mode), speed == 25 ? "25G" : "10G", 3);
	for (i = 0, j = 0; i < MAX_PORTS; i++) {
		if (mask&(1<<i)) {
			printf("Port%-2d%-23s : %s\n", j, mode,
				   result&(1<<i) ? "Up" : "Down");
			j++;
		}
	}
	snprintf_s_s(path+offset, sizeof(path)-offset, "/%s", "pkvl_a_version");
	get_sysfs_attr(path, version, sizeof(status));
	printf("%-29s : %s\n", "Retimer A Version", version);
	snprintf_s_s(path+offset, sizeof(path)-offset, "/%s", "pkvl_b_version");
	get_sysfs_attr(path, version, sizeof(status));
	printf("%-29s : %s\n", "Retimer B Version", version);
}

static void print_phy_info(fpga_properties props, struct _phy_config *config)
{
	fpgainfo_print_common("//****** PHY ******//", props);
	if ((config->group_num == 0) || (config->group_num == -1)) {
		print_phy_group_info(props, 0);
	}
	if ((config->group_num == 1) || (config->group_num == -1)) {
		print_phy_group_info(props, 1);
	}
	if (config->show_pkvl == 1) {
		print_pkvl_info(props);
	}
}

#define PHY_GETOPT_STRING ":G:h"
int parse_phy_args(int argc, char *argv[])
{
	struct option longopts[] = {
		{"group", required_argument, NULL, 'G'},
		{"help", no_argument, NULL, 'h'},
		{0, 0, 0, 0},
	};
	int getopt_ret;
	int option_index;

	/* default configuration */
	phy_config.group_num = -1;

	optind = 0;
	while (-1 != (getopt_ret = getopt_long(argc, argv, PHY_GETOPT_STRING,
										   longopts, &option_index))) {
		const char *tmp_optarg = optarg;

		if (optarg && ('=' == *tmp_optarg)) {
			++tmp_optarg;
		}

		switch (getopt_ret) {
		case 'G':
            if (NULL == tmp_optarg)
                break;
			if (!strcmp("0", tmp_optarg)) {
				phy_config.group_num = 0;
			} else if (!strcmp("1", tmp_optarg)) {
				phy_config.group_num = 1;
			} else if (!strcmp("all", tmp_optarg)) {
				phy_config.group_num = -1;
			} else {
				fprintf(stderr, "Invalid argument '%s' of option group\n",
						tmp_optarg);
				return -1;
			}
			break;

		case 'h':   /* help */
			phy_help();
			return -1;

		case ':':   /* missing option argument */
			fprintf(stderr, "Missing option argument\n");
			phy_help();
			return -1;

		case '?':
		default:    /* invalid option */
			fprintf(stderr, "Invalid cmdline options\n");
			phy_help();
			return -1;
		}
	}

	return 0;
}

fpga_result phy_filter(fpga_properties *filter, int argc, char *argv[])
{
	fpga_result res = FPGA_INVALID_PARAM;

	if (0 == parse_phy_args(argc, argv)) {
		res = fpgaPropertiesSetObjectType(*filter, FPGA_DEVICE);
		fpgainfo_print_err("setting type to FPGA_DEVICE", res);
	}
	return res;
}

fpga_result phy_command(fpga_token *tokens, int num_tokens, int argc,
						char *argv[])
{
	(void)argc;
	(void)argv;
	fpga_result res = FPGA_OK;
	fpga_properties props;

	int i = 0;
	for (i = 0; i < num_tokens; ++i) {
		res = fpgaGetProperties(tokens[i], &props);
		ON_FPGAINFO_ERR_GOTO(res, out_destroy, "reading properties from token");

		print_phy_info(props, &phy_config);
		fpgaDestroyProperties(&props);
	}

	return res;

out_destroy:
	fpgaDestroyProperties(&props);
	return res;
}
